# The statistical analysis of the potassium chloride pH in the soil naturally
# divided into the agricuture and forest part, based on data collected on both
# locations during the time.

# The R code is a part of the paper

# A permutation approach to the analysis of spatio-temporal data in the presence
# of heteroscedasticity
# Authors: Veronika Rimalova, Alessandra Menafoglio, Alessia Pini, Vilem Pechanec,
#          Eva Fiserova 
# Contact address: veronika.rimalova@upol.cz
# submitted to Envirometrics, July 25, 2019

# The code contains 
#   1. data preparation: raw data are turned into functional observations using 
#         B-splines basis. The data are further smoothed using the penalised 
#         residual sum of squares criterion  via the generalised cross-validation
#   2. analysis of spatial dependence using the variogram
#   3. spatial-temporal model - the ordinary least-squares approach: model, 
#         testing homoscedasticity
#   4. spatial-temporal model - the weighted least-squares approach: model, testing
#         homoscedasticity, testing the effect of covariates
#   5. Model quality evaluation

##################################################################################



library(MASS)
library(readxl)
library(fda)
library(fields)
library(R.matlab)
library(Rcpp)
library(ggplot2)
library(gstat)
library(fdagstat)
library(Rfast)


#################################################################################
# 1. functional observations  
#################################################################################

kcl.raw <- read_excel("soil measurements.xlsx",sheet='pH KCl')

# remove the line 'year'
kcl.raw=kcl.raw[-9,]

fdnames.pH=list('Month', 'Distance', 'pH')
basis = create.bspline.basis(rangeval=c(3,10), nbasis=10, norder=4)
lambda=1

kcl.functions = smooth.basis(argvals=seq(3,10), y=as.matrix(kcl.raw[,-1]), 
                             fdPar(basis,2,10^lambda), fdnames=fdnames.pH)$fd


# evaluate the functions - vectors of length 120
eval.kcl.functions=as.data.frame(eval.fd(seq(from=3,to=10,length.out=120),kcl.functions))


# exclude the ecotone - only field and forest observations remain
eval.kcl=eval.kcl.functions[,c(1:5,7:16,18:22)]


########################################################################################
# 2. examining the spatial correlation through variography
########################################################################################

# define coordinates
# distinguish field and forest data - by 0 and 1
coord = cbind.data.frame(rep(c(rep(0,5),rep(1,5)),2),c(rep(c(-15,-12,-9,-6,-3,3,6,9,12,15),2)))
names(coord) = c('dummy','dist')
coord=as.data.frame(coord)

fvar <- fstat(NULL, 'vgram', coord, eval.kcl)

fvar<- estimateDrift("~dummy", fvar, .type = "OLS")

fvar <- fvariogram(formula      = "~dist" ,
                   fvar                   ,   # fstat structure.
                   Nlags        = 10  ,   # number of lags.
                   LagMax       = 12   ,   # maximum lag distance.
                   directions   = "omni",   # it computes empirical variograms along each input direction.
                   LagTolerance = 10 ,   # it will adjust to lag half.
                   ArgStep      = 1    ,   # sampling step of the functions.
                   angTolerance = 8,       # angular tolerance for directional variogram computation.
                   useResidual  = T)

print(plotVariogram(fvar, ggReturn=TRUE) + coord_cartesian(ylim=c(0,10)))
fvar=fitVariograms(fvar,model=vgm(5,model='Nug'))
print(plotVariogram(fvar, ggReturn=TRUE) + coord_cartesian(ylim=c(0,10)))

# nugget effect - no spatial dependence among observations

# further we model the data as spatially independent, i.e. 
# diagonal structure of the variance-covariance matrix


##################################################################################
# 3. functional regression model - ordinary least-squares
##################################################################################

number.of.observations=20
number.of.permutations=1000

# rearrange the data into groups - field and forest
field=eval.kcl[,c(1:5,11:15)]
forest=eval.kcl[,c(6:10,16:20)]
kcl.data=cbind(field,forest)


# prepare the design matrix X
intercept=c(rep(1,20))
soil=c(c(rep(0,number.of.observations/2),rep(1,number.of.observations/2))) 

X=matrix(nrow=number.of.observations,ncol=2,data=c(intercept,soil))


# precompute matrices for speeding up the computations
projection.ols=solve(t(X)  %*% X) %*% t(X)
betas=projection.ols %*% t(kcl.data)

# estimate variances separately for field and forest
residuals=t(kcl.data)- X %*% betas

first.sample.size=number.of.observations/2
second.sample.size=number.of.observations/2

# field variance
field.residuals=residuals[c(1:10),]
forest.residuals=residuals[c(11:20),]

norm.field=numeric(first.sample.size)
for (line in 1:(first.sample.size)) {
  norm.field[line]=(norm((field.residuals[line,]), type="2"))^2
  
}
var.field=sum(norm.field)/(first.sample.size)




# forest variance

norm.forest=numeric(second.sample.size)
for (line in 1:(second.sample.size)) {
  norm.forest[line]=(norm((forest.residuals[line,]), type="2"))^2
  
}
var.forest=sum(norm.forest)/(second.sample.size)




# 3.1 test for equality of variances

# H0: sigma^2_field = sigma^2_forest

# define the test statistic for equality of variances
T.var=abs(log(var.field/var.forest))


#Permutation test
T.var.perm=numeric(number.of.permutations)
for(perm in 1:number.of.permutations){
  
  permutation <- sample(number.of.observations)
  
  permuted.residuals=residuals[permutation,]
  
  
  permuted.field=permuted.residuals[c(1:10),]
  permuted.forest=permuted.residuals[c(11:20),]
  
  permuted.norm.field=numeric(first.sample.size)
  for (line in 1:(first.sample.size)) {
    permuted.norm.field[line]=(norm((permuted.field[line,]), type="2"))^2
    
  }
  permuted.var.field=sum(permuted.norm.field)/(first.sample.size)
  
  
  
  
  permuted.norm.forest=numeric(second.sample.size)
  for (line in 1:(second.sample.size)) {
    permuted.norm.forest[line]=(norm((permuted.forest[line,]), type="2"))^2
    
  }
  permuted.var.forest=sum(permuted.norm.forest)/(second.sample.size)
  
  
  
  T.var.perm[perm]=abs(log(permuted.var.field/permuted.var.forest))
  
  
}

# proportion of all permutations leading to higher value of the test statistic 
# than the original data's test statistic
(p.value.T.var=sum(T.var.perm>T.var)/number.of.permutations)


# we reject the null hypothesis ->  weighted least-squares model for the data

################################################################################
# 4. functional regression model - the weighted least-squares approach
#################################################################################

# define the weight matrix W
W=diag(c(rep(1/var.field,first.sample.size),rep(1/var.forest,second.sample.size)))

# predefine matrices for faster computation
projection.wls=solve(t(X)  %*% W   %*% X) %*% t(X) %*% W

sd.field=sqrt(var.field)
sd.forest=sqrt(var.forest)

sd.matrix=matrix(c(rep(sd.field,first.sample.size),rep(sd.forest,second.sample.size)),120,
                 number.of.observations,byrow=T)


#wls estimate of parameters beta_0, beta_1 from the model
betas.wls=projection.wls %*% t(kcl.data)

# estimate the residuals from wls model
residuals.wls=t(kcl.data)- X %*% betas.wls

# divide residuals by corresponding standard deviation - exchangeable residuals
divided.residuals.wls=(residuals.wls)/(t(sd.matrix)) 




# 4.1 Test for the effect of covariate 'soil'
# H0: beta1=0

# Under H0, consider the reduced model

# design matrix
X.reduced=c(rep(1,number.of.observations))

# to speed up computations
projection.wls.reduced=solve(t(X.reduced) %*% W %*% X.reduced) %*% t(X.reduced) %*% W


# define the test statistic

C1=matrix(ncol(X)-1,1,data=0)
C2=diag(1,ncol(X)-1)
C=cbind(C1,C2)

part.of.stat.wls=solve(C%*%solve(t(X)%*%W%*%X)%*%t(C))

#test statistic for observed data (full model)
T0.from.data=trapzc(step=1,t(C%*%betas.wls)%*%part.of.stat.wls%*%(C%*%betas.wls))

#wls estimate of parameter beta0 from the reduced model
reduced.beta0= projection.wls.reduced %*% t(kcl.data)

# put beta0 from reduced model into matrix
reduced.beta0.matrix=matrix(reduced.beta0,120,number.of.observations)

# compute residuals of the reduced model
reduced.residuals=kcl.data-reduced.beta0.matrix

#divide residuals by (true) standard deviation
reduced.residuals.divided=reduced.residuals/sd.matrix



# Permutation test 
T_H0.reduced=numeric(number.of.permutations) 

for(perm in 1:number.of.permutations){
  
  permutation <- sample(number.of.observations)
  
  #sum of reduced responses and permuted residuals (multiplied by standard deviation)
  permuted.responses=reduced.beta0.matrix+sd.matrix*reduced.residuals.divided[,permutation]
  permuted.responses=as.data.frame(permuted.responses)
  
  #estimate parameters of the full model from permuted responses
  beta.from.permuted=projection.wls %*% t(permuted.responses)
  
  
  #compute test statistic for the full model estimated from permuted responses 
  T_H0.reduced[perm]=trapzc(step=1,t(C%*%beta.from.permuted)%*%part.of.stat.wls%*%(C%*%beta.from.permuted))
  
  
}

#proportion of all permutations leading to higher value of the test statistic than the original data's test statistic
(p.value=sum(T_H0.reduced>T0.from.data)/number.of.permutations)

# reject H0 -> covariate 'soil' is significant


###############################################################################
# 5. Model quality evaluation
###############################################################################

# 5.1 R-squared
# RSS
residual.sum.of.squares=numeric(number.of.observations)
for (line in 1:number.of.observations) {
  residual.sum.of.squares[line]=(norm((residuals.wls[line,]), type="2"))^2
  
}

#TSS
matrix.of.mean=t(matrix(rowmeans(as.matrix(kcl.data)),120,number.of.observations))
total.variance=t(kcl.data)-matrix.of.mean

total.sum.of.squares=numeric(number.of.observations)
for (line in 1:number.of.observations) {
  total.sum.of.squares[line]=(norm((total.variance[line,]), type="2"))^2
  
}


(R.squared=1-sum(residual.sum.of.squares)/sum(total.sum.of.squares))

# NRMSE
MSE=sum(residual.sum.of.squares)/(number.of.observations)
(RMSE=sqrt(MSE))

mean.curve=rowmeans(as.matrix(kcl.data))
(NRMSE=RMSE/norm(mean.curve,type='2'))
